#include "pqtree.h"

#include <algorithm>
#include <assert.h>

void
pqnode::compute_total_weight()
{
	weight = 0;
	for (pqnode * c = first; c; c = c->right) {
		weight += c->weight;
	}
}

void
pqnode::compute_side_weight_perm()
{
	side_weight = 0;
	side_first = 0;

	double best_end = 0;
	pqnode *best_n = 0;

	for (pqnode *c = first; c; c = c->right) {
		pqnode *t = c;

		// new side children
		double gain = c->side_weight - std::max(c->weight, 0.0);
		if (gain > best_end) {
			best_end = gain;
			t = best_n;
			best_n = c;
		}

		if (t && t->weight > 0) {
			side_weight += t->weight;
			t->side_next = side_first;
			side_first = t;
		}

	}

	// add the best side children to the beginning
	if (best_n) {
		side_weight += best_n->side_weight;
		best_n->side_next = side_first;
		side_first = best_n;
	}
}

void
pqnode::compute_best_inner_child()
{
	double best = 0;
	pqnode *bestn = 0;

	for (pqnode *c = first; c; c = c->right) {
		if (c->inner_weight > best) {
			best = c->inner_weight;
			bestn = c;
		}
	}

	if (best > inner_weight) {
		inner_weight = best;
		bestn->inner_next = 0;
		inner_first = bestn;
	}
}


void
pqnode::compute_inner_weight_perm()
{
	inner_weight = 0;
	inner_first = 0;

	double best_end[2] = {0, 0};
	pqnode *best_n[2] = {0, 0};

	for (pqnode *c = first; c; c = c->right) {
		pqnode *t = c;
		double gain = c->side_weight - std::max(c->weight, 0.0);

		// new side children
		if (gain > best_end[0]) {
			best_end[1] = best_end[0];
			t = best_n[1];
			best_n[1] = best_n[0];
			best_end[0] = gain;
			best_n[0] = c;
		}
		else if (gain > best_end[1]) {
			best_end[1] = gain;
			t = best_n[1];
			best_n[1] = c;
		}

		if (t && t->weight > 0) {
			inner_weight += t->weight;
			t->inner_next = inner_first;
			inner_first = t;
		}
	}


	// add the best side children to the beginning
	if (best_n[0]) {
		inner_weight += best_n[0]->side_weight;
		best_n[0]->inner_next = inner_first;
		inner_first = best_n[0];
	}
	if (best_n[1]) {
		inner_weight += best_n[1]->side_weight;
		best_n[1]->inner_next = inner_first;
		inner_first = best_n[1];
	}

	compute_best_inner_child();
}

void
pqnode::compute_side_weight_queue()
{
	side_weight = 0;
	side_first = 0;

	double best_left = 0;
	pqnode *best_leftn = 0;

	double cur = 0;
	for (pqnode *c = first; c; c = c->right) {
		if (cur + c->side_weight > best_left) {
			best_left = cur + c->side_weight;
			best_leftn = c;
		}
		cur += c->weight;
	}

	double best_right = 0;
	pqnode *best_rightn = 0;

	cur = 0;
	for (pqnode *c = last; c; c = c->left) {
		if (cur + c->side_weight > best_right) {
			best_right = cur + c->side_weight;
			best_rightn = c;
		}
		cur += c->weight;
	}

	if (best_left > best_right && best_leftn) {
		side_weight = best_leftn->side_weight;
		for (pqnode *c = best_leftn->left; c; c = c->left) {
			side_weight += c->weight;
			c->side_next = side_first;
			side_first = c;
		}
		best_leftn->side_next = side_first;
		side_first = best_leftn;
	}
	else if (best_rightn) {
		side_weight = best_rightn->side_weight;
		for (pqnode *c = best_rightn->right; c; c = c->right) {
			side_weight += c->weight;
			c->side_next = side_first;
			side_first = c;
		}
		best_rightn->side_next = side_first;
		side_first = best_rightn;
	}
}

void
pqnode::compute_inner_weight_queue()
{
	inner_weight = 0;
	inner_first = 0;

	double best_left = 0;
	pqnode *best_leftn = first;

	double best = 0;
	pqnode *best_last = 0;
	pqnode *best_first = 0;


	double cur = 0;

	for (pqnode *c = first; c; c = c->right) {
		if (c->side_weight + cur + best_left > best) {
			best = c->side_weight + cur + best_left;
			best_last = c;
			best_first = best_leftn;
		}

		cur += c->weight;
		if (c->side_weight - cur > best_left) {
			best_left = c->side_weight - cur;
			best_leftn = c;
		}
	}

	if (best_first == 0) {
		compute_best_inner_child();
		return;
	}

	for (pqnode *c = best_first->right; c && c != best_last; c = c->right) {
		inner_weight += c->weight;
		c->inner_next = inner_first;
		inner_first = c;
	}

	inner_weight += best_first->side_weight;
	best_first->inner_next = inner_first;
	inner_first = best_first;

	if (best_last != best_first) {
		inner_weight += best_last->side_weight;
		best_last->inner_next = inner_first;
		inner_first = best_last;
	}

	compute_best_inner_child();
}

void
pqnode::assign_interval_inner(boolvector & out)
{
    if (type == pqtype_leaf) {
        out[id] = 1;
        return;
	}

    pqnode *c = inner_first;

    if (c == 0) return;

    if (c->inner_next == 0) {
        c->assign_interval_inner(out);
        return;
	}

    c->assign_interval_side(out);
    c = c->inner_next;
    c->assign_interval_side(out);
    c = c->inner_next;
    while (c) {
        c->assign_interval_full(out);
        c = c->inner_next;
	}
}

void
pqnode::assign_interval_side(boolvector & out)
{
    if (type == pqtype_leaf) {
        out[id] = 1;
        return;
	}

	pqnode *c = side_first;

	if (c == 0) return;

	c->assign_interval_side(out);
	for (c = c->side_next; c; c = c->side_next) {
		c->assign_interval_full(out);
	}
}


void
pqnode::assign_interval_full(boolvector & out)
{
    if (type == pqtype_leaf) {
        out[id] = 1;
        return;
	}
	for (pqnode *c = first; c; c = c->right) {
		c->assign_interval_full(out);
	}
}

void
pqtree::assign_optimal(boolvector & out)
{
    m_sentinel->first->assign_interval_inner(out);
}


double
pqtree::find_max()
{
    reset_counter();
    m_queue.reset();

	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		m_queue.push(&m_leaves[i]);
	}


	while (!m_queue.empty()) {
		pqnode *n = m_queue.pop();

		if (n->type != pqtype_leaf) {
			n->compute_total_weight();
			if (n->type == pqtype_p) {
				n->compute_side_weight_perm();
				n->compute_inner_weight_perm();
			}
			else if (n->type == pqtype_q) {
				n->compute_side_weight_queue();
				n->compute_inner_weight_queue();
			}
		}
		else {
			n->side_weight = n->weight;
			n->inner_weight = n->weight;
		}


		pqnode *p = n->parent;
		assert(p);
		p->visited_count++;
		if (p->visited_count == p->child_count && p != m_sentinel)
			m_queue.push(p);
	}


	return m_sentinel->first->inner_weight;
}

void
pqtree::set_weights(doublevector & weights)
{
	for (uint32_t i = 0; i < m_leaves.size(); i++) {
		m_leaves[i].weight = weights[i];
		//printf("%lf ", weights[i]);
	}
	//printf("\n");
}
